@@ -41,16 +41,18 @@
 (defvar nethack-status-header-line-format)
 (defvar nethack-status-mode-line-format)
 (defvar nethack-status-buffer-format)
-(defvar nethack-message-style)
 (defvar nethack-use-tiles)
 (defvar nethack-version)
 (defvar nethack-tile-vector)
 (defvar nethack-proc)
 (defvar nethack-lisprec-record)
+(defvar nethack-want-completing-read)
+(defvar nethack-tiles-scale)
 (declare-function nethack-map-mode "nethack")
 (declare-function nethack-status-mode "nethack")
 (declare-function nethack-message-mode "nethack")
 (declare-function nethack-send "nethack")
+(declare-function nethack-send-and-wait "nethack")
 
 
 ;;; Buffer handling
@@ -368,6 +370,8 @@ Values taken from
 http://fileformats.archiveteam.org/wiki/DEC_Special_Graphics_Character_Set,
 accessed 2021-04-23.")
 
+;; set to nil to disable the check
+(defvar nethack-nhapi-print-glyph--previous-ch t)
 (defun nethack-nhapi-print-glyph (x y color glyph tile ch &optional attr)
   "Insert glyph into `nethack-map-buffer'."
   (set-buffer nethack-map-buffer)
@@ -375,10 +379,12 @@ accessed 2021-04-23.")
   (let ((inhibit-read-only t))
     (if (and nethack-use-tiles (display-images-p))
         (save-excursion
-          (let ((buffer-read-only nil))
+          (let ((buffer-read-only nil)
+                (image (elt nethack-tile-vector tile)))
             (goto-char (gamegrid-cell-offset x y))
             (delete-char 1)
-            (insert-image (elt nethack-tile-vector tile))))
+            (image--set-property image :scale nethack-tiles-scale)
+            (insert-image image)))
       (cond
         ((or (nethack-options-set-p 'DECgraphics) (string-match-p "^DECgraphics$" (or (nethack-options-set-p 'symset) "")))
          (nethack-gamegrid-set-cell
@@ -393,7 +399,21 @@ accessed 2021-04-23.")
                 ch))))
         ((or (nethack-options-set-p 'IBMgraphics) (string-match-p "^IBMgraphics\\(?:_1\\|_2\\)?$" (or (nethack-options-set-p 'symset) "")))
          (nethack-gamegrid-set-cell x y (decode-char 'cp437 ch)))
-        (t (nethack-gamegrid-set-cell x y ch)))
+        (t
+         ;; try to infer symset by checking the character used for the horizontal wall.
+         (when nethack-nhapi-print-glyph--previous-ch
+           (if (eq nethack-nhapi-print-glyph--previous-ch ch)
+               (when-let ((symset (pcase ch
+                                    (241 "DECgraphics")
+                                    (196 "IBMgraphics")
+                                    (?- (setq nethack-nhapi-print-glyph--previous-ch nil))
+                                    (_ nil))))
+                 (setq nethack-nhapi-print-glyph--previous-ch nil)
+                 (push `("symset" ,symset) nethack-options)
+                 (warn "Inferred symset to be %s, please put OPTIONS=symset:%1$s in your .nethackrc" symset))
+             (setq nethack-nhapi-print-glyph--previous-ch ch)))
+
+         (nethack-gamegrid-set-cell x y ch)))
       (set-text-properties (gamegrid-cell-offset x y)
                            (1+ (gamegrid-cell-offset x y))
                            `(face
@@ -401,6 +421,79 @@ accessed 2021-04-23.")
                              glyph
                              ,glyph)))))
 
+
+(defun nethack-completing-read-filter (action)
+  (let* ((polymorph nil) ;; TODO
+         (all-categories '("Coins" "Weapons" "Armor" "Amulets" "Potions" "Scrolls" "Comestibles" "Spellbooks" "Rings" "Wands" "Tools" "Gems/Stones"))
+         (action-to-category-preds `(("drink" . (("Potions")))
+                                     ("adjust" . (,all-categories))
+                                     ("use or apply" . (,all-categories))
+                                     ("wear" . (("Armor") . ("being worn")))
+                                     ("take off" . (("Armor") . (t "being worn")))
+                                     ("drop" . (,all-categories . ("being worn")))
+                                     ("dip" . (,all-categories))
+                                     ("dip into" . (("Potions")))
+                                     ("eat" . (,@(pcase polymorph
+                                                  ('gelatinous-cube `(,all-categories))
+                                                  ('metalivore `(,all-categories))
+                                                  (_ '(("Comestibles"))))))
+                                     ("write with" . (("Coins" "Weapons" "Miscellaneous" "Wands" "Tools" "Gems/Stones")))
+                                     ("wield" . ((,@all-categories "Miscellaneous") . ("being worn" "wielded")))
+                                     ("read" . (("Scrolls" "Spellbooks")))
+                                     ("ready" . ((,all-categories) . ("being worn")))
+                                     ("throw" . (,all-categories . ("being worn")))
+                                     ("fire" . (,all-categories . ("being worn")))
+                                     ("put on" . (("Amulets" "Rings" "Armor") . ("being worn")))
+                                     ("remove" . (("Amulets" "Rings" "Armor") . (t "being worn")))))
+         (category-predicates (cdr (assoc-string action action-to-category-preds)))
+         (inventory (append (when (string= action "write with") '(("-" "your fingertip" . "Miscellaneous")))
+                            (when (string= action "wield") '(("-" "your bare hands" . "Miscellaneous")))
+                            nethack--inventory)))
+    (lambda (str _pred flag)
+      (pcase flag
+        ('metadata
+         (list 'metadata
+               (cons 'display-sort-function #'identity)
+               (cons 'annotation-function
+                     (lambda (cand)
+                       (format #("    %s" 4 6 (face font-lock-doc-face))
+                               (substring-no-properties (cadr (assoc-string cand inventory))))))
+               (cons 'group-function
+                     (lambda (cand transform)
+                       (if transform cand
+                         (cddr (assoc-string (substring-no-properties cand 0 1) inventory)))))))
+        ('t
+         (if (string-blank-p str)
+             (all-completions str (cl-remove-if-not
+                (lambda (i)
+                  (let* ((name (cadr i))
+                         (category (cddr i))
+                         ;; yes, if you name a weapon "(wielded)" or some dumb shit like that, this will break.
+                         (name-matcher (lambda (s) (string-match-p (format "(%s)$" (regexp-quote s)) name))))
+                    (and (member category (car category-predicates))
+                         (if (eq (cadr category-predicates) t) ;; invert match
+                             (cl-some name-matcher (cddr category-predicates))
+                           (not (cl-some name-matcher (cdr category-predicates)))))))
+                inventory))
+           (all-completions
+            str
+            (lambda (s _ _)
+              (mapcar
+               #'car
+               (cl-remove-if-not
+                (lambda (i)
+                  (let* ((name (cadr i))
+                         (category (cddr i))
+                         (name-matcher (lambda (s) (string-match-p (format "(%s)$" (regexp-quote s)) name))))
+                    (when (and (member category (car category-predicates))
+                         (if (eq (cadr category-predicates) t) ;; invert match
+                             (cl-some name-matcher (cddr category-predicates))
+                           (not (cl-some name-matcher (cdr category-predicates)))))
+                    (or
+                     (string-match-p (regexp-quote s) (car i))
+                     (string-match-p (regexp-quote s) (cadr i))))))
+                inventory))))))))))
+
 (defun nethack-nhapi-yn-function (ques choices default)
   (if nethack-proc
       (let (key)
@@ -411,21 +504,38 @@ accessed 2021-04-23.")
         (when (/= default 0)
           (setq choices (cons default choices)))
 
-        ;; Add some special keys of our own to the choices
-        (setq choices (cons 13 choices))
+        (if (and nethack-want-completing-read
+                 (string-match "\\(What do you want to \\(dip \\(.*\\) into\\|.*\\)\\? \\)\\[" ques))
+            (let* ((prompt (match-string 1 ques))
+                   (action (if (match-string 3 ques) "dip into" (match-string 2 ques)))
+                   (ret (condition-case _ (completing-read prompt (nethack-completing-read-filter action)) (quit "\0"))))
+              (while (not (or (assoc-string ret nethack--inventory) (string= ret "\0")))
+                (run-with-timer 0.01 nil (lambda () (minibuffer-message "match required")))
+                (setq ret (condition-case _ (completing-read prompt (nethack-completing-read-filter action)) (quit "\0"))))
+              (if (and (not (string= ret "\0")) (numberp current-prefix-arg) (> current-prefix-arg 0))
+                  (let* ((inhibit-quit nil)
+                         (nstr (number-to-string current-prefix-arg))
+                         (msd (substring nstr 0 1))
+                         (remaining-count (string-to-number (substring nstr 1))))
+                    (nethack-send-and-wait (string-to-char msd))
+                    (funcall (keymap-lookup nethack-map-mode-map ret) remaining-count))
+                (nethack-send (string-to-char ret))))
 
-        (setq key (nethack-read-char (concat ques " ")))
-        (when (> (length choices) 1)
-          (while (not (member key choices))
-            (setq key (nethack-read-char (concat
-                                          (format "(bad %d) " key)
-                                          ques " ")))))
-        ;; 13, 27, and 7 are abort keys
-        (nethack-send (if (or (= 13 key)
-                              (= 27 key)
-                              (= 7 key))
-                          default
-                        key)))
+          ;; Add some special keys of our own to the choices
+          (setq choices (cons 13 choices))
+
+          (setq key (nethack-read-char (concat ques " ")))
+          (when (> (length choices) 1)
+            (while (not (member key choices))
+              (setq key (nethack-read-char (concat
+                                            (format "(bad %d) " key)
+                                            ques " ")))))
+          ;; 13, 27, and 7 are abort keys
+          (nethack-send (if (or (= 13 key)
+                                (= 27 key)
+                                (= 7 key))
+                            default
+                          key))))
     (message ques)))
 
 (defun nethack-nhapi-ask-direction (prompt)
@@ -530,25 +640,20 @@ Do not edit the value of this variable.  Instead, change the value of
 
 (defvar nethack-inventory-need-update nil
   "If non-nil, at the next command prompt, update the menu.")
+(defvar nethack--inventory nil)
 
 (defun nethack-nhapi-update-inventory ()
+  (setq nethack--inventory nil)
   (setq nethack-inventory-need-update t))
 
 (defun nethack-nhapi-doprev-message ()
-  (cl-case nethack-message-style
-    (:map
-     (nethack-clear-message)
-     (nethack-message 'atr-none nethack-last-message))
-    (t
-     (save-selected-window
-       (save-current-buffer             ; is this redundant since we
-                                        ; already save the selected
-                                        ; window? -rcy
-         (walk-windows (lambda (w)
-                         (select-window w)
-                         (set-buffer (window-buffer))
-                         (when (eq (current-buffer) nethack-message-buffer)
-                           (scroll-down)))))))))
+  (save-selected-window
+    (save-current-buffer
+      (walk-windows (lambda (w)
+                      (select-window w)
+                      (set-buffer (window-buffer))
+                      (when (eq (current-buffer) nethack-message-buffer)
+                        (scroll-down)))))))
 
 (defun nethack-nhapi-update-positionbar (_features))
 
@@ -560,6 +665,8 @@ Do not edit the value of this variable.  Instead, change the value of
   (setq nethack-directory (file-name-directory executable))
   (when (and (nethack-options-set-p 'tiled_map) (null nethack-use-tiles))
     (message "You have OPTIONS=tiled_map set in your nethackrc; consider setting nethack-use-tiles"))
+  (setq nethack-nhapi-print-glyph--previous-ch t)
+  (setq nethack--inventory nil)
   ;; clean up old buffers
   (mapc (lambda (b) (kill-buffer (cdr b))) nethack-menu-buffer-table)
   (setq nethack-menu-buffer-table nil)
@@ -572,23 +679,15 @@ Do not edit the value of this variable.  Instead, change the value of
 
 (defun nethack-nhapi-create-message-window ()
   "Create the message buffer."
-  (cl-case nethack-message-style
-    (:map
-     ;; we need to create this buffer because messages come in before
-     ;; the map is set up.
-     (with-current-buffer (get-buffer-create "*nethack map*")
-       (let ((inhibit-read-only t))
-         (insert (make-string nethack-map-width 32) "\n"))))
-    (t
-     (with-current-buffer (get-buffer-create "*nethack message*")
-       (nethack-message-mode)
-       (let ((inhibit-read-only t))
-         (erase-buffer))
-       (setq nethack-message-highlight-overlay
-             (make-overlay (point-max) (point-max)))
-       (overlay-put nethack-message-highlight-overlay
-                    'face 'nethack-message-highlight-face)
-       (setq nethack-message-buffer (current-buffer))))))
+  (with-current-buffer (get-buffer-create "*nethack message*")
+    (nethack-message-mode)
+    (let ((inhibit-read-only t))
+      (erase-buffer))
+    (setq nethack-message-highlight-overlay
+          (make-overlay (point-max) (point-max)))
+    (overlay-put nethack-message-highlight-overlay
+                 'face 'nethack-message-highlight-face)
+    (setq nethack-message-buffer (current-buffer))))
 
 (defun nethack-nhapi-create-status-window ()
   "Create the status buffer."
@@ -694,11 +793,12 @@ The TYPE argument is legacy and serves no real purpose."
   "Major mode for Nethack menus.
 
 \\{nethack-menu-mode-map}"
-  (setq mode-name (string-join `("NetHack Menu" ,(when how (symbol-name how))) " "))
+  (kill-all-local-variables)
+  (setq mode-name (string-join `("NetHack Menu" ,@(when how (list (symbol-name how)))) " "))
   (setq major-mode #'nethack-menu-mode)
-  (use-local-map nethack-menu-mode-map)
   (setq nethack-menu-how how)
-  (run-hooks 'nethack-menu-mode-hook))
+  (use-local-map nethack-menu-mode-map)
+  (run-mode-hooks 'nethack-menu-mode-hook))
 
 (defun nethack-menu-toggle-item (&optional count)
   "Toggle the menu item that is associated with the key event that
@@ -736,18 +836,139 @@ was actually toggled."
         (message "No such menu option: %c" last-command-event)
         (goto-char old-point)))))
 
-(defun nethack-menu-toggle-all-items ()
-  "Toggle all menu items, only for pick-any menus."
+(defun nethack-menu-toggle-items (&optional behavior page-only)
+  "Toggle items in a pick-any menu buffer.
+
+Items are inverted by default, and selected/deselected if behavior is
+`select'/`deselect'.
+
+Set page-only to t to only toggle visible menu items."
   (interactive)
+  (when (eq nethack-menu-how 'pick-any)
+    (save-excursion
+      (let ((inhibit-read-only t))
+        (goto-char (if page-only (window-start) (point-min)))
+        (while (re-search-forward nethack-menu-item-regexp (when page-only (window-end)) t)
+          (let ((value (match-string 2)))
+            (if (and (or (string-equal value "-")
+                         (eq behavior 'select))
+                     (not (eq behavior 'deselect)))
+                (replace-match "+" nil nil nil 2)
+              (replace-match "-" nil nil nil 2))))))))
+
+(defun nethack-menu-select-page ()
+  "Select all visible menu items."
+  (interactive)
+  (nethack-menu-toggle-items 'select t))
+
+(defun nethack-menu-select-all ()
+  "Select all menu items."
+  (interactive)
+  (nethack-menu-toggle-items 'select))
+
+(defun nethack-menu-deselect-page ()
+  "Deselect all visible menu items."
+  (interactive)
+  (nethack-menu-toggle-items 'deselect t))
+
+(defun nethack-menu-deselect-all ()
+  "Deselect all menu items."
+  (interactive)
+  (nethack-menu-toggle-items 'deselect))
+
+(defun nethack-menu-invert-page ()
+  "Invert selection of all visible menu items."
+  (interactive)
+  (nethack-menu-toggle-items nil t))
+
+(defun nethack-menu-invert-all ()
+  "Invert selection of all menu items."
+  (interactive)
+  (nethack-menu-toggle-items))
+
+(defun nethack-menu-toggle-category (category)
+  "Toggle all items under CATEGORY in a pick-any menu buffer.
+
+If CATEGORY cannot be found, fallback to `nethack-menu-toggle-item'."
+  (when (catch 'fallback
+          (save-excursion
+            (let ((inhibit-read-only t))
+              (goto-char (point-min))
+              (unless (and (eq nethack-menu-how 'pick-any)
+                           (re-search-forward (format "^%s$" category) nil t))
+                (throw 'fallback t))
+              (beginning-of-line 2)
+              (while (looking-at nethack-menu-item-regexp)
+                (let ((value (match-string 2)))
+                  (if (string-equal value "-")
+                      (replace-match "+" nil nil nil 2)
+                    (replace-match "-" nil nil nil 2)))
+                (beginning-of-line 2)))))
+    (nethack-menu-toggle-item)))
+
+;; kinda useless as only menu item in coins already has accelerator $
+(defun nethack-menu-toggle-coins ()
+  (interactive)
+  (nethack-menu-toggle-category "Coins"))
+
+(defun nethack-menu-toggle-weapons ()
+  (interactive)
+  (nethack-menu-toggle-category "Weapons"))
+
+(defun nethack-menu-toggle-armor ()
+  (interactive)
+  (nethack-menu-toggle-category "Armor"))
+
+(defun nethack-menu-toggle-amulets ()
+  (interactive)
+  (nethack-menu-toggle-category "Amulets"))
+
+(defun nethack-menu-toggle-scrolls ()
+  (interactive)
+  (nethack-menu-toggle-category "Scrolls"))
+
+(defun nethack-menu-toggle-potions ()
+  (interactive)
+  (nethack-menu-toggle-category "Potions"))
+
+(defun nethack-menu-toggle-comestibles ()
+  (interactive)
+  (nethack-menu-toggle-category "Comestibles"))
+
+(defun nethack-menu-toggle-spellbooks ()
+  (interactive)
+  (nethack-menu-toggle-category "Spellbooks"))
+
+(defun nethack-menu-toggle-rings ()
+  (interactive)
+  (nethack-menu-toggle-category "Rings"))
+
+(defun nethack-menu-toggle-wands ()
+  (interactive)
+  (nethack-menu-toggle-category "Wands"))
+
+(defun nethack-menu-toggle-tools ()
+  (interactive)
+  (nethack-menu-toggle-category "Tools"))
+
+(defun nethack-menu-toggle-gems ()
+  (interactive)
+  (nethack-menu-toggle-category "Gems/Stones"))
+
+(defun nethack-menu-search-toggle (regexp)
+  "Toggle all menu items matching REGEXP."
+  (interactive "MSearch for: ")
   (when (eq nethack-menu-how 'pick-any)
     (save-excursion
       (let ((inhibit-read-only t))
         (goto-char (point-min))
         (while (re-search-forward nethack-menu-item-regexp nil t)
-          (let ((value (match-string 2)))
-            (if (string-equal value "-")
-                (replace-match "+" nil nil nil 2)
-              (replace-match "-" nil nil nil 2))))))))
+          (let ((item (match-string 3))
+                (value (match-string 2)))
+            (when (string-match-p regexp item)
+              (if (string-equal value "-")
+                  (replace-match "+" nil nil nil 2)
+                (replace-match "-" nil nil nil 2)))))))))
 
 (defun nethack-menu-goto-next ()
   "Move to the next selectable menu item."
@@ -847,6 +1068,11 @@ buffer."
       (insert-char ?\n 1 nil)
       (when (nethack-options-set-p 'menucolors)
         (nethack-options-highlight-menu))
+      (when nethack-inventory-need-update
+        (if (= accelerator -1)
+            (push `(,str) nethack--inventory)
+          (let ((start (save-excursion (search-backward str))))
+            (push (cons (char-to-string accelerator) (buffer-substring start (+ start (length str)))) (cdar nethack--inventory)))))
       (run-hooks 'nethack-add-menu-hook))))
 
 ;; FIXME: xemacs propertize bug here
@@ -869,6 +1095,14 @@ the menu is dismissed."
     (if nethack-inventory-need-update
         (progn
           (setq nethack-inventory-need-update nil)
+          ;; transpose from '((CATEGORY (ACCEL . NAME))) to '((ACCEL NAME . CATEGORY))
+          (setq nethack--inventory
+                (reverse (mapcan
+                          (lambda (category)
+                            (mapcar (lambda (item)
+                                      `(,(car item) ,(cdr item) . ,(car category)))
+                                    (cdr category)))
+                          nethack--inventory)))
           (nethack-send nil))
       (progn
         (unless nethack-active-menu-buffer
@@ -1017,6 +1251,7 @@ the menu is dismissed."
   (setq nethack-options-mon-polycontrol mon-polycontrol))
 
 (defun nethack-nhapi-need-options-file ()
+  (setq nethack-options nil)
   (if (and (file-exists-p nethack-options-file)
            (not nethack-lisprec-record)
            ;; remote nethack sessions use the shell, so the command is longer
