@@ -5,8 +5,8 @@
 ;; Author: Ryan Yeske <rcyeske@vcn.bc.ca>
 ;; Maintainer: George Huebner <george@feyor.sh>
 ;; Created: Sat Mar 18 11:31:52 2000
-;; Package-Version: 20251013.2341
-;; Package-Revision: 4213dd9c6941
+;; Package-Version: 20251025.1205
+;; Package-Revision: c3d4b8980fa8
 ;; Keywords: games
 ;; URL: https://github.com/Feyorsh/nethack-el
 
@@ -108,6 +108,11 @@ Valid values are :map, :header-line, :mode-line, or t."
   :type '(boolean)
   :group 'nethack)
 
+(defcustom nethack-want-completing-read t
+  "When non-nil, prefer using `completing-read' to get input."
+  :type '(boolean)
+  :group 'nethack)
+
 ;;; Insert variables that control how the status gets displayed here.
 
 (defcustom nethack-use-tiles nil
@@ -131,14 +136,6 @@ Valid values are :map, :header-line, :mode-line, or t."
   :type '(hook)
   :group 'nethack)
 
-(defcustom nethack-message-style t
-  "Decides where messages appear.
-:map means messages display in the map buffer.
-t means in a separate buffer."
-  :type '(symbol)
-  :options '(:map t)
-  :group 'nethack)
-
 (defcustom nethack-end-hook nil
   "Hook run when nethack has ended."
   :type '(hook)
@@ -748,6 +745,35 @@ Assumes nethack is not already running."
     (with-current-buffer nethack-map-buffer (nethack-map-mode)))
   (nethack-command-redraw-screen 2))
 
+(defvar nethack-tiles-scale nil)
+(defun nethack-rescale-tiles ()
+  (save-excursion
+    (with-current-buffer nethack-map-buffer
+      (goto-char (point-min))
+      (while (not (eobp))
+        (when-let ((image (get-text-property (point) 'display)))
+          (image--set-property image :scale nethack-tiles-scale))
+        (goto-char (next-single-char-property-change (point) 'display))))
+    (clear-image-cache)))
+
+(defun nethack-text-scale-increase (inc)
+  (interactive "p")
+  (unless (derived-mode-p 'nethack-map-mode)
+    (error "Not in NetHack map mode"))
+  (if (eq (next-single-char-property-change (point-min) 'display) (point-max))
+      (text-scale-increase inc)
+    (setq nethack-tiles-scale (min 8.0 (ffloor (+ nethack-tiles-scale inc))))
+    (nethack-rescale-tiles)))
+
+(defun nethack-text-scale-decrease (dec)
+  (interactive "p")
+  (unless (derived-mode-p 'nethack-map-mode)
+    (error "Not in NetHack map mode"))
+  (if (eq (next-single-char-property-change (point-min) 'display) (point-max))
+      (text-scale-decrease dec)
+    (setq nethack-tiles-scale (max 0.5 (- nethack-tiles-scale dec)))
+    (nethack-rescale-tiles)))
+
 ;;;; Process code to communicate with the Nethack executable
 (defconst nethack-prompt-regexp
   "^\\(command\\|menu\\|dummy\\|direction\\|number\\|string\\)> *")
@@ -811,13 +837,13 @@ PROC is the process object and MSG is the exit message."
     (when raw-print-buffer
       (switch-to-buffer raw-print-buffer))))
 
-(defvar nethack-log-process-text t)
 (defun nethack-log (string &optional print-timestamp)
-  (when nethack-log-process-text
-    (with-current-buffer (get-buffer-create nethack-log-buffer)
-      (goto-char (point-max))
-      (when print-timestamp
-        (insert (format "%s ; %s\n" (string-trim string) (time-subtract (current-time) nethack-start-time)))))))
+  (with-current-buffer (get-buffer-create nethack-log-buffer)
+    (goto-char (point-max))
+    (insert (string-trim string))
+    (when print-timestamp
+      (insert (format " ; %s" (time-subtract (current-time) nethack-start-time))))
+    (insert "\n")))
 
 (defvar nethack-at-prompt nil)
 (defvar nethack-at-prompt-hook nil
@@ -899,7 +925,7 @@ delete the contents, perhaps logging the text."
   (setq-local scroll-margin 3)
   (setq-local cursor-in-non-selected-windows nil)
   (when (and nethack-use-tiles (display-graphic-p))
-    (let ((cookie (face-remap-add-relative 'default :height 16)))
+    (let ((cookie (face-remap-add-relative 'default :height 8)))
       (setq nethack-use-tiles--font-height (or nethack-use-tiles--font-height cookie)))
     ;; for not clobbering our keybindings, and preventing the user
     ;; from accidentally messing up the map tiles
